﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace ModbusDemo.windows
{
    class Modbuslib
    {
        [DllImport(@"LibModbus.dll", EntryPoint = "fnLibModbus", CharSet = CharSet.Unicode)]
        public static extern uint fnLibModbus();
        [DllImport(@"LibModbus.dll", EntryPoint = "fnLibModbus485", CharSet = CharSet.Unicode)]
        public static extern int fnLibModbus485(byte ucPort,int ulBaudRate,int timeout, ref uint handle);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnReadCoilStatus", CharSet = CharSet.Unicode)]
        public static extern int fnReadCoilStatus(uint handle,byte m_usSlaveAddress,ushort address, ushort num, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.Byte[] dest, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnReadInputStatus")]
        public static extern int fnReadInputStatus(uint handle, byte m_usSlaveAddress, ushort address, ushort num, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.Char[] dest, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnReadHoldingReg", CharSet = CharSet.Unicode)]
        public static extern int fnReadHoldingReg(uint handle, byte m_usSlaveAddress, ushort address, ushort num, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.UInt16[] dest, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnReadInputReg")]
        public static extern int fnReadInputReg(uint handle, byte m_usSlaveAddress, ushort address, ushort num, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.UInt16[] dest, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnWriteSingleCoil")]
        public static extern int fnWriteSingleCoil(uint handle, byte m_usSlaveAddress, ushort address, ushort state, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnWriteHoldingReg")]
        public static extern int fnWriteHoldingReg(uint handle, byte m_usSlaveAddress, ushort address, ushort state, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnWriteMultipleCoils")]
        public static extern int fnWriteMultipleCoils(uint handle, byte m_usSlaveAddress, ushort address, ushort num, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.Char[] src, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnWriteMultipleRegs")]
        public static extern int fnWriteMultipleRegs(uint handle, byte m_usSlaveAddress, ushort address, ushort num, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.UInt16[] src, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnReportSlaveID")]
        public static extern int fnReportSlaveID(uint handle, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.Char[] src, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnMaskWriteReg")]
        public static extern int fnMaskWriteReg(uint handle, byte m_usSlaveAddress, ushort address, ushort amask, ushort omask, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnReadWriteMultipleRegs")]
        public static extern int fnReadWriteMultipleRegs(uint handle, byte m_usSlaveAddress, ushort waddress, ushort wnum, ushort raddress, ushort rnum, [MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 1)] System.UInt16[] dest, byte[] Ip, ushort wTcpPort);
        [DllImport(@"LibModbus.dll", EntryPoint = "fnLibModbusClose")]
        public static extern void fnLibModbusClose(uint handle);

        private uint hModbusHandle = 0;
        private System.UInt16[] dest = new System.UInt16[10] { 0, 0, 3, 4, 5, 6, 7, 8, 9, 0 };

        private byte mport;
        private int mrate;
        private int mtimeout;

        public Modbuslib()
        {
            this.hModbusHandle = fnLibModbus();
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="port">端口</param>
        /// <param name="rate">波特率</param>
        /// <param name="timeout">超时设置</param>
        public Modbuslib(byte port, int rate, int timeout)
        {
            this.mport = port;
            this.mrate = rate;
            this.mtimeout = timeout;
        }

        public int ConnectModbus()
        {
            int ret;
            if (hModbusHandle != 0)
            {
                fnLibModbusClose(hModbusHandle);

            }
            ret = fnLibModbus485(mport, mrate, mtimeout, ref hModbusHandle);
            return ret;
        }

        ~Modbuslib()
        {
            fnLibModbusClose(hModbusHandle);
        }

        /// <summary>
        /// 读线圈状态
        /// </summary>
        /// <param name="addr">地址</param>
        /// <param name="num">线圈数目</param>
        /// <param name="ip">设备IP</param>
        /// <param name="wport">端口号 默认502</param>
        /// <param name="buf">读取结果</param>
        /// <returns>
        /// 0 -- no error
        /// 1 -- illegal register address.
        /// 2 -- illegal argument
        /// 3 -- porting layer error
        /// 4 -- insufficient resources
        /// 5 -- I/O error
        /// 6 -- protocol stack in illegal state
        /// 7 -- timeout error occurred
        /// 8 -- slave address error
        /// 9 -- exception error occurred
        /// 10-- CRC check error
        /// 11-- send data error
        /// 12-- receive data error
        /// 13-- TCP connect error
        /// 14-- close error
        /// </returns>
        public int ReadCoilStatus(byte SlaveAdd,ushort addr, ushort num, String ip, ushort wport, ref byte[] buf)
        {
            int ret;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnReadCoilStatus(hModbusHandle, SlaveAdd,(ushort)(addr - 1), num, buf, bIp, wport);
            return ret;
        }

        public int ReadInputStatus(byte SlaveAdd, ushort addr, ushort num, String ip, ushort wport, ref char[] buf)
        {
            int ret;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnReadInputStatus(hModbusHandle, SlaveAdd, (ushort)(addr - 1), 1, buf, bIp, 502);
            return ret;
        }

        public int ReadHoldingReg(byte SlaveAdd, ushort addr, ushort num, String ip, ushort wport, ref UInt16[] buf)
        {
            int ret;
            byte[] bIp=null;
            if (ip != null)
            {
                bIp = Encoding.Unicode.GetBytes(ip);
            }
            
            ret = fnReadHoldingReg(hModbusHandle, SlaveAdd,(ushort)(addr - 1), num, buf, bIp, wport);
            return ret;
        }

        public int ReadInputReg(byte SlaveAdd, ushort addr, ushort num, String ip, ushort wport, ref UInt16[] buf)
        {
            int ret;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnReadInputReg(hModbusHandle, SlaveAdd,(ushort)(addr - 1), 1, dest, bIp, wport);
            buf[0] = dest[0];
            return ret;
        }

        public int WriteSingleCoil(byte SlaveAdd,ushort addr, ushort state, String ip, ushort wport)
        {
            int ret = 0;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnWriteSingleCoil(hModbusHandle, SlaveAdd,(ushort)(addr - 1), state, bIp, wport);
            return ret;
        }

        public int WriteHoldingReg(byte SlaveAdd, ushort addr, ushort state, String ip, ushort wport)
        {
            int ret = 0;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnWriteHoldingReg(hModbusHandle, SlaveAdd,(ushort)(addr - 1), state, bIp, wport);
            return ret;
        }

        //public int ReadHoldingReg(byte SlaveAdd, ushort addr, ushort num, String ip, ushort wport, ref char[] buf)
        //{
        //    int ret;
        //    byte[] bIp = Encoding.Unicode.GetBytes(ip);
        //    ret = fnWriteMultipleCoils(hModbusHandle, SlaveAdd,(ushort)(addr - 1), 1, buf, bIp, wport);
        //    return ret;
        //}

        public int WriteMultipleRegs(byte SlaveAdd, ushort addr, ushort num, String ip, ushort wport, ref ushort[] buf)
        {
            int ret;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnWriteMultipleRegs(hModbusHandle, SlaveAdd,(ushort)(addr - 1), num, buf, bIp, wport);
            return ret;
        }

        public int ReportSlaveID(String ip, ushort wport, ref char[] buf)
        {
            int ret;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnReportSlaveID(hModbusHandle, buf, bIp, 502);
            return ret;
        }

    

        public int MaskWriteReg(byte SlaveAdd, ushort addr, ushort amask, ushort omask, String ip, ushort wport)
        {
            int ret = 0;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnMaskWriteReg(hModbusHandle, SlaveAdd,(ushort)(addr - 1), amask, omask, bIp, wport);
            return ret;
        }

        public int ReadWriteMultipleRegs(byte SlaveAdd, ushort waddr, ushort wnum, ushort raddr, ushort rnum, String ip, ushort wport, ref UInt16[] buf)
        {
            int ret = 0;
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            ret = fnReadWriteMultipleRegs(hModbusHandle, SlaveAdd,(ushort)(waddr - 1), wnum, raddr, rnum, buf, bIp, wnum);
            return ret;
        }

        ///////////////////////////////////////////////////////////////////////////

        public byte GetSlaveIDandName(ref UInt16[] names)
        {
            int ret;
            byte Slaveid= 1;
            String ip = "127.0.0.1";
            //UInt16[] buf = new ushort[2];
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
            int i = 0;
            for (i = Slaveid; i < 255; i++ )
            {
                ret = fnReadHoldingReg(hModbusHandle, (byte)i, (ushort)(211 - 1), 2, names, bIp, 502);

                if(ret == 0)
                {
                    Slaveid = (byte)i;
                    break;
                }
            }

            if(i >255)
            {
                Slaveid = 0;
            }



            return Slaveid;
        }


        public String GetVersion(Byte slaveid)
        {
            int ret;

            String version = "0.0";
            String ip = "127.0.0.1";
            UInt16[] buf = new ushort[2];
            byte[] bIp = Encoding.Unicode.GetBytes(ip);
          //  int i = 0;
            ret = fnReadHoldingReg(hModbusHandle, slaveid, (ushort)(213 - 1), 2, buf, bIp, 502);

            if(ret == 0)
            {
                version = buf[0].ToString() + "." + buf[1].ToString();
            }
            return version;
        }

    }
}
